#include "Arduino.h"
#include <stdbool.h>
#include "StepperDM556T.h"
#include "N64_Controller.h"

/*
   This code is the combination of stepper driver code and the controller reading. How to use:
   Wire the arduino according to the new wiring diagram: Wiring_Diagram_ArtStudio_v2.pdf
   Plug in the Arduino either into a computer or a 12V power supply.

   There is a safety button implemented on the back of the controller to make sure the motors don't move
   accidentally. Press and hold the Z button on the back of the remote to enable the motors

   If this is undesired, please change the ENABLE_SAFETY_BUTTON value to 0 to remove that feature.

   To move the picture on the wall either Left, Right, Up or Down, there are two modes of operation:
   
   1) Simply press the controllers DPAD the corresponding direction. 
   The steppers will move X_STEPS_AT_A_TIME to the left or right after the corresponding Left/Right button is pressed on the DPAD. 
   The steppers will move Y_STEPS_AT_A_TIME up or down after the corresponding Up/Down button is pressed on the DPAD. 

   2) To continuously move the photo, use the joystick to move. As long as the joystick is pressed it will move at MOTOR_SPEED
   in the direction of the joyscik being pressed.


   Connecting to the serial monitor for arduino will show what the system is doing for more context if things are not
   working properly. Setting the DISPLAY_CONTROLLER_DEBUG_INFO to 1 will show if the controller is operating properly or not.

   Reach out with any questions or things you would like changed!

   Thank you for your business and I hope this helps your future work!

   Andrew
*/
// Things to change:
// How far the test code moves each stepper.

#define MOTOR_SPEED (5)  // The delay time in ms between pulses, Range is 1 (very fast) to 100 (very slow)
#define ENABLE_SAFETY_BUTTON (1) //change to 0 if the safety button is not desired.
#define DISPLAY_CONTROLLER_DEBUG_INFO (0) // change to 1 if you want to see the controller data output
#define X_STEPS_AT_A_TIME (200)  // How many steps the motors should move each time the joystick is pressed in a direction
#define Y_STEPS_AT_A_TIME (400)  // How many steps the motors should move each time the joystick is pressed in a direction
#define JOYSTICK_LIMIT (20) // value between 1 - 60 for how far you have to push the joystick for a register. 1 is barely at all, 60 is almost to the edge.
//End things to change



#define REAR_BUTTON_BIT 2
#define LEFT_DPAD_BIT 6
#define RIGHT_DPAD_BIT 7
#define UP_DPAD_BIT 4
#define DOWN_DPAD_BIT 5
// Pins for the X stepper motor controller
#define X_MOTOR_ENABLE_PIN 11
#define X_MOTOR_DIRECTION_PIN 10
#define X_MOTOR_PULSE_PIN 9
#define X_MOTOR_LEFT_DIRECTION COUNTERCLOCKWISE
#define X_MOTOR_RIGHT_DIRECTION CLOCKWISE

// Pins for the Y1 stepper motor controller
#define LEFT_Y_MOTOR_ENABLE_PIN 8
#define LEFT_Y_MOTOR_DIRECTION_PIN 7
#define LEFT_Y_MOTOR_PULSE_PIN 6
#define LEFT_Y_MOTOR_UP_DIRECTION COUNTERCLOCKWISE
#define LEFT_Y_MOTOR_DOWN_DIRECTION CLOCKWISE

// Pins for the Y2 stepper motor controller
#define RIGHT_Y_MOTOR_ENABLE_PIN 5
#define RIGHT_Y_MOTOR_DIRECTION_PIN 4
#define RIGHT_Y_MOTOR_PULSE_PIN 3
#define RIGHT_Y_MOTOR_UP_DIRECTION CLOCKWISE
#define RIGHT_Y_MOTOR_DOWN_DIRECTION COUNTERCLOCKWISE
// END changeable code

typedef enum wd {UP, DOWN, LEFT, RIGHT} WALL_DIRECTION;

// Objects for the stepper motor control
StepperDM556T x_stepper;
StepperDM556T right_y_stepper;
StepperDM556T left_y_stepper;

void setup()
{
  Serial.begin(9600);
  setup_controller();

  // Setup the objects for the stepper motor controls
  right_y_stepper.init(RIGHT_Y_MOTOR_DIRECTION_PIN, RIGHT_Y_MOTOR_PULSE_PIN, RIGHT_Y_MOTOR_ENABLE_PIN);
  left_y_stepper.init(LEFT_Y_MOTOR_DIRECTION_PIN, LEFT_Y_MOTOR_PULSE_PIN, LEFT_Y_MOTOR_ENABLE_PIN);
  left_y_stepper.enable();
  right_y_stepper.enable();
  x_stepper.init(X_MOTOR_DIRECTION_PIN, X_MOTOR_PULSE_PIN, X_MOTOR_ENABLE_PIN);
  x_stepper.disable();
  Serial.println("Welcome! Please move the joystick left, right, up or down. v7");
}

void move_motors_distance(uint32_t steps, WALL_DIRECTION dir);



//Debounce values for the controller:
BoolDebounce left_dpad;
BoolDebounce right_dpad;
BoolDebounce up_dpad;
BoolDebounce down_dpad;

BoolDebounce left_joystick;
BoolDebounce right_joystick;
BoolDebounce up_joystick;
BoolDebounce down_joystick;

void loop()
{
  
  int i = 0;
  unsigned char data = 0;
  unsigned char addr = 0;;

  // Command to send to the gamecube
  // The last bit is rumble, flip it to rumble
  // yes this does need to be inside the loop, the
  // array gets mutilated when it goes through N64_send
  unsigned char command[] = {0x01};
  // don't want interrupts getting in the way
  noInterrupts();
  // send those 3 bytes
  N64_send(command, 1);
  // read in data and dump it to N64_raw_dump
  N64_get();
  // end of time sensitive code
  interrupts();

  // translate the data in N64_raw_dump to something useful
  translate_raw_data();

  if (DISPLAY_CONTROLLER_DEBUG_INFO) {
    for (i = 0; i < 16; i++) {
      Serial.print(N64_raw_dump[i], DEC);
    }
    Serial.print(' ');
    Serial.print(N64_status.stick_x, DEC);
    Serial.print(' ');
    Serial.print(N64_status.stick_y, DEC);
    Serial.print(" \n");
  }
  

  //Safety button to be sure not accidentally moved. Must be pressed and held to keep the system safe.
  if (!ENABLE_SAFETY_BUTTON || N64_raw_dump[REAR_BUTTON_BIT])
  {
    //determine what to do with joystick.
    if (right_dpad.has_value_rose())
    {
      move_motors_distance(X_STEPS_AT_A_TIME, RIGHT);
    }
    else if (left_dpad.has_value_rose())
    {
      move_motors_distance(X_STEPS_AT_A_TIME, LEFT);
    }
    else if (up_dpad.has_value_rose())
    {
      move_motors_distance(Y_STEPS_AT_A_TIME, UP);
    }
    else if (down_dpad.has_value_rose())
    {
      move_motors_distance(Y_STEPS_AT_A_TIME, DOWN);
    }

    //Check the joy stick value here:
    else if (left_joystick.has_value_rose())
    {
      move_motors_distance(10, LEFT);
    }
    else if (right_joystick.has_value_rose())
    {
      move_motors_distance(10, RIGHT);
    }
    //If the motor is moving and the joystick is still held left OR right, we will keep adding steps onto the motor turning.
    else if (!x_stepper.is_movement_complete() && (right_joystick.is_value_high() || left_joystick.is_value_high())){
      x_stepper.set_steps_remaining(3);
    }
    else if (up_joystick.has_value_rose())
    {
      move_motors_distance(10, UP);
    }
    else if (down_joystick.has_value_rose())
    {
      move_motors_distance(10, DOWN);
    }
    //If either of the motors are moving and the joystick is still held up OR down, we will keep adding steps onto the motor turning.
    else if ((!right_y_stepper.is_movement_complete() || !left_y_stepper.is_movement_complete()) && (down_joystick.is_value_high() || up_joystick.is_value_high())){
      left_y_stepper.set_steps_remaining(3);
      right_y_stepper.set_steps_remaining(3);
    }



    //We only do something special for the x motor when it is done moving.
    //once the x stepper is done moving, we shut off to not heat up excessively.
    //if the safety button is enabled, then we can turn on the motor when the button is pressed. Otherwise we turn it off on idle.
    if (ENABLE_SAFETY_BUTTON) {
      if (!x_stepper.is_enabled()){
          x_stepper.enable();
      }
    }
    else if (x_stepper.is_enabled() && x_stepper.is_movement_complete())
    {
      x_stepper.disable();
    }
  
    //Update the debouncers to watch for the rising edge of the signals from the controller. 
    //This is used to make sure the motors are started only once
    left_dpad.update_value(N64_raw_dump[LEFT_DPAD_BIT]);
    right_dpad.update_value(N64_raw_dump[RIGHT_DPAD_BIT]);
    up_dpad.update_value(N64_raw_dump[UP_DPAD_BIT]);
    down_dpad.update_value(N64_raw_dump[DOWN_DPAD_BIT]);
    left_joystick.update_value(N64_status.stick_x < -JOYSTICK_LIMIT); //Use the comparison against 5 to make it a boolean signal to debounce.
    right_joystick.update_value(N64_status.stick_x > JOYSTICK_LIMIT);
    up_joystick.update_value(N64_status.stick_y > JOYSTICK_LIMIT);
    down_joystick.update_value(N64_status.stick_y < -JOYSTICK_LIMIT);
    
    // Each time through the loop update the stepper's movements
    right_y_stepper.update_movement();
    left_y_stepper.update_movement();
    x_stepper.update_movement();
  }
  else
  {
    //Disable all the steppers.
    x_stepper.disable();
    left_y_stepper.set_steps_remaining(0);
    right_y_stepper.set_steps_remaining(0);
  }
}

//Only starts movement if the motors are done moving.
void move_motors_distance(uint32_t steps, WALL_DIRECTION dir)
{
    //determine what to do with joystick.
    switch (dir)
    {
      case RIGHT:
        if (x_stepper.is_movement_complete()) {
          Serial.println("Moving Right");
          if (!x_stepper.is_enabled()){
            x_stepper.enable();
          }
          x_stepper.start_movement(X_MOTOR_RIGHT_DIRECTION, steps, MOTOR_SPEED);
        }
        break;
      case LEFT:
        if (x_stepper.is_movement_complete()) {
          Serial.println("Moving Left");
          if (!x_stepper.is_enabled()){
            x_stepper.enable();
          }
          x_stepper.start_movement(X_MOTOR_LEFT_DIRECTION, steps, MOTOR_SPEED);
        }
        break;
      case UP:
        if (right_y_stepper.is_movement_complete()) {
          Serial.println("Moving Up");
          right_y_stepper.start_movement(RIGHT_Y_MOTOR_UP_DIRECTION, steps, MOTOR_SPEED);
        }
        if (left_y_stepper.is_movement_complete()) {
          left_y_stepper.start_movement(LEFT_Y_MOTOR_UP_DIRECTION, steps, MOTOR_SPEED);
        }
         break;
      case DOWN:
        if (right_y_stepper.is_movement_complete()) {
          Serial.println("Moving Down");
          right_y_stepper.start_movement(RIGHT_Y_MOTOR_DOWN_DIRECTION, steps, MOTOR_SPEED);
        }
        if (left_y_stepper.is_movement_complete()) {
          left_y_stepper.start_movement(LEFT_Y_MOTOR_DOWN_DIRECTION, steps, MOTOR_SPEED);
        }
        break;
    }
  }


